from django.conf import settings
from django.db.models import Q, Prefetch, prefetch_related_objects, Count
from rest_framework.exceptions import ValidationError, NotFound
from rest_framework.generics import (
    ListAPIView, CreateAPIView, RetrieveAPIView, UpdateAPIView, DestroyAPIView,
    ListCreateAPIView, RetrieveUpdateDestroyAPIView
)
from rest_framework import status
from rest_framework.response import Response
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import (
    Corpus, Element, ElementPath, Right, MetaData, AllowedMetaData,
    Classification, ClassificationState, Transcription
)
from arkindex.documents.serializers.elements import (
    CorpusSerializer, ElementSerializer, ElementSlimSerializer, ElementListSerializer,
    ElementCreateSerializer, ElementNeighborsSerializer, ElementParentSerializer,
    MetaDataUpdateSerializer,
)
from arkindex.project.openapi import AutoSchema
from arkindex.documents.serializers.light import CorpusAllowedMetaDataSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.project.mixins import CorpusACLMixin, DeprecatedMixin
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly, IsAuthenticated
from arkindex.project.triggers import corpus_delete
from arkindex.project.tools import BulkMap

from uuid import UUID
import operator


classifications_queryset = Classification.objects.select_related('ml_class', 'source').order_by('-confidence')

best_classifications_prefetch = Prefetch(
    'classifications',
    queryset=classifications_queryset
    # Keep best or validated distinct classifications
    .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True))
    # Remove rejected classifications
    .exclude(state=ClassificationState.Rejected).distinct(),
    to_attr='best_classes'
)


def _fetch_children_count(elements):
    """
    From a list of elements, sets the `children_count` attribute to a boolean
    indicating if an element has children elements.

    The initial goal was to have a way for the frontend to tell if an element had any children;
    having a count was optional. It is possible to just use an Exists subquery since Django 2.2:
    Element.objects.annotate(has_children=Exists(ElementPath.objects.filter(path__last=OuterRef('pk'))))
    This works but makes the element list queries 20 times slower, due to the way the Postgres query planner
    handles this.

    Instead, we use a second query only on element paths with a GROUP BY, then re-link the results back to
    Element instances after Django REST Framework does the pagination. This only makes the query twice as slow,
    and this is made optional in a `with_children_count` GET parameter. The difference in execution time
    between telling if an element has children and telling how many children an element has becomes negligible.

    The ElementsListMixin uses arkindex.project.tools.BulkMap to apply this method and perform the second request
    *after* DRF's pagination, because there is no way to perform post-processing after pagination in Django
    without having to use Django private methods.
    """
    counts = dict(
        ElementPath.objects
                   .filter(path__last__in=map(operator.attrgetter('id'), elements))
                   .values('path__last')
                   .annotate(count=Count('element_id', distinct=True))
                   .values_list('path__last', 'count')
    )

    for element in elements:
        element.children_count = counts.get(element.id, 0)

    return elements


class ElementsListMixin(object):
    """
    Provides common features across element list endpoints (list, parents, children)
    """
    serializer_class = ElementListSerializer
    queryset = Element.objects.all()
    openapi_overrides = {
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter elements by type',
                'required': False,
                'schema': {
                    'type': 'string',
                }
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'best_class',
                'in': 'query',
                'description': 'Restrict to or exclude elements with best classes, '
                               'or restrict to elements with a specific best class',
                'required': False,
                'schema': {'anyOf': [
                    {'type': 'string', 'format': 'uuid'},
                    {'type': 'boolean'},
                ]}
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False,
                }
            },
            {
                'name': 'with_children_count',
                'in': 'query',
                'description': (
                    'Include the `children_count` integer to count direct children elements of each element. '
                    'Otherwise, `children_count` will always be null.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
            {
                'name': 'with_transcription_sources_count',
                'in': 'query',
                'description': (
                    'Include the `transcription_sources_count` '
                    'integer to count available transcriptions sources for each element. '
                    'Otherwise, `transcription_sources_count` will always be null.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            }
        ]
    }

    def get_filters(self):
        filters = {
            'corpus__in': Corpus.objects.readable(self.request.user),
        }
        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.request.query_params['name']

        if 'type' in self.request.query_params:
            filters['type__slug'] = self.request.query_params['type']

        only_folder = self.request.query_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        class_filter = self.request.query_params.get('best_class')
        if class_filter is not None:
            best_classification_queryset = Classification \
                .objects \
                .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True)) \
                .exclude(state=ClassificationState.Rejected)

            try:
                class_filter = UUID(class_filter)
            except (TypeError, ValueError):
                # Assume it is a boolean
                # Exclusion is handled in get_excludes
                if class_filter.lower() not in ('false', '0'):
                    filters['classifications__in'] = best_classification_queryset
            else:
                filters['classifications__in'] = best_classification_queryset.filter(ml_class_id=class_filter)

        return filters

    def get_excludes(self):
        excludes = {}

        class_filter = self.request.query_params.get('best_class')
        if class_filter is not None:
            try:
                UUID(class_filter)
            except (TypeError, ValueError):
                if class_filter.lower() in ('false', '0'):
                    # Exclude all best classes
                    excludes['classifications__in'] = Classification \
                        .objects \
                        .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True)) \
                        .exclude(state=ClassificationState.Rejected)

        return excludes

    def get_prefetch(self):
        prefetch = {'corpus', 'zone__image__server', 'type'}

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            prefetch.add(best_classifications_prefetch)

        return prefetch

    def get_order_by(self):
        return ('corpus', 'type__slug', 'name', 'id')

    @property
    def apply_distinct(self):
        # Use queryset.distinct() whenever best_class is defined
        return bool(self.request.query_params.get('best_class'))

    def filter_queryset(self, queryset):
        # Ordering by ID is required by postgres to order elements with common corpus, type and name
        queryset = queryset \
            .filter(**self.get_filters()) \
            .exclude(**self.get_excludes()) \
            .prefetch_related(*self.get_prefetch()) \
            .order_by(*self.get_order_by())
        if self.apply_distinct:
            queryset = queryset.distinct()

        transcription_sources_count = self.request.query_params.get('with_transcription_sources_count')
        if transcription_sources_count and transcription_sources_count.lower() not in ('false', '0'):
            queryset = queryset.annotate(transcription_sources_count=Count('transcriptions__source_id', distinct=True))

        with_children_count = self.request.query_params.get('with_children_count')
        if with_children_count and with_children_count.lower() not in ('false', '0'):
            queryset = BulkMap(_fetch_children_count, queryset)

        return queryset


class ElementsList(ElementsListMixin, CorpusACLMixin, ListAPIView):
    """
    List elements and filter by type, name, corpus, ML class
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElements',
        'description': 'List elements and filter by type, name, corpus, ML class',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'corpus',
                'in': 'query',
                'description': 'Filter elements by corpus ID',
                'required': False,
                'schema': {'type': 'string', 'format': 'uuid'},
            },
        ]
    })

    def get_filters(self):
        filters = super().get_filters()

        if 'corpus' in self.request.query_params:
            try:
                corpus_id = UUID(self.request.query_params['corpus'])
            except ValueError:
                raise ValidationError({'corpus': ['Not a valid uuid']})
            filters['corpus'] = self.get_corpus(corpus_id)

        return filters


class ElementParents(ElementsListMixin, ListAPIView):
    """
    List all parents of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementParents',
        'description': 'List all parents of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (parents, grandparents, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    @property
    def is_recursive(self):
        recursive_param = self.request.query_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    @property
    def apply_distinct(self):
        return super().apply_distinct or not self.is_recursive

    def get_filters(self):
        filters = super().get_filters()

        if not self.is_recursive:
            # List direct parents only: elements whose IDs are the last in the element's paths
            filters['id__in'] = ElementPath.objects \
                .filter(element_id=self.kwargs['pk']) \
                .values('path__last')

        return filters

    def filter_queryset(self, queryset):
        if not self.is_recursive:
            return super().filter_queryset(queryset)

        # Redefine the queryset here as get_ascending needs its filters as an argument
        # and returns a list instead of an actual queryset
        parents = Element.objects.get_ascending(self.kwargs['pk'], **self.get_filters())

        # The list is unsorted; prefetch using prefetch_related_objects and sort with sorted
        if self.apply_distinct:
            # Apply .distinct() using a dict
            parents = list({parent.id: parent for parent in parents}.values())

        prefetch_related_objects(parents, *self.get_prefetch())

        with_children_count = self.request.query_params.get('with_children_count')
        if with_children_count and with_children_count.lower() not in ('false', '0'):
            _fetch_children_count(parents)

        return sorted(parents, key=operator.attrgetter('corpus', 'type.slug', 'name', 'id'))


class ElementChildren(ElementsListMixin, ListAPIView):
    """
    List all children of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementChildren',
        'description': 'List all children of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (children, grandchildren, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    def get_filters(self):
        filters = super().get_filters()
        recursive_param = self.request.query_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # Only list direct children
            # We could use Element.objects.filter(paths__path__last=volid) directly,
            # but it is 100× slower due to the implementation of path__last.
            # get_descending uses 'path @> ARRAY[element.id]', which is much faster,
            # so we append the __last filter to it.
            filters['paths__path__last'] = self.kwargs['pk']

        return filters

    def get_order_by(self):
        return ('paths__ordering', )

    def get_queryset(self):
        return Element.objects.get_descending(self.kwargs['pk'])


class RelatedElementsList(DeprecatedMixin, ListAPIView):
    """
    List all parents and children of a single element
    """
    serializer_class = ElementListSerializer
    openapi_overrides = {
        'operationId': 'ListRelatedElements',
        'security': [],
        'tags': ['elements']
    }
    deprecation_message = \
        'This endpoint has been deprecated. ' \
        'Use ListElementParents and ListElementChildren instead to get parent and child elements separately.'


class ElementRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        return Element.objects \
            .filter(corpus__in=corpora) \
            .select_related(
                'corpus',
                'type',
                'zone__image__server',
            ) \
            .prefetch_related(
                'metadatas__revision__repo',
                'metadatas__entity',
                Prefetch('classifications', queryset=classifications_queryset)
            )

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        rights = obj.corpus.get_acl_rights(request.user)
        if request.method == 'DELETE' and Right.Admin not in rights:
            self.permission_denied(request, message='You do not have admin access to this element.')
        elif request.method != 'GET' and Right.Write not in rights:
            self.permission_denied(request, message='You do not have write access to this element.')

    def perform_destroy(self, instance):
        children_count = ElementPath.objects.filter(path__contains=[instance.id]).count()
        if children_count:
            raise ValidationError({
                'element': "Element '{}' is linked to {} children elements".format(instance.id, children_count)
            })
        super().perform_destroy(instance)

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET':
            context['element'] = self.get_object()
        return context


class ElementNeighbors(ListAPIView):
    """
    List neighboring elements
    """
    serializer_class = ElementNeighborsSerializer
    openapi_overrides = {
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'n',
                'in': 'query',
                'description': 'Number of neighbors to retrieve around the element',
                'required': False,
                'schema': {
                    'type': 'integer',
                    'minimum': 1,
                    'maximum': 10,
                },
            },
        ]
    }

    def get_queryset(self):
        n = self.request.query_params.get('n', 1)
        try:
            n = int(n)
        except (TypeError, ValueError):
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})
        if not 1 <= n <= 10:
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})

        try:
            element = Element.objects.get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound

        return Element.objects.get_neighbors(element, n)


class NoRequestBodySchema(AutoSchema):
    def _get_request_body(self, path, method):
        # Remove request body from OpenAPI schema
        return None


class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )
    schema = NoRequestBodySchema()
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)


class ManageSelection(ListAPIView):
    """
    Manage all selected elements
    """
    serializer_class = ElementListSerializer
    permission_classes = (IsAuthenticated, )
    openapi_overrides = {
        'operationId': 'ManageSelection',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            }
        ]
    }

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])
        if self.request.session.get('elements_selected') is None:
            self.request.session['elements_selected'] = []

    def get_queryset(self):
        elements = self.request.session.get('elements_selected')
        filtered_queryset = Element.objects.filter(id__in=elements).order_by('corpus', 'type', 'name')

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            return filtered_queryset.prefetch_related(best_classifications_prefetch)

        return filtered_queryset

    def post(self, request, *args, **kwargs):
        if 'ids' not in request.data:
            return Response(status=status.HTTP_400_BAD_REQUEST)

        elements = Element.objects.filter(id__in=request.data['ids']) \
            .prefetch_related('corpus', 'zone__image__server', 'type')
        if elements.count() != len(set(request.data['ids'])):
            return Response(status=status.HTTP_404_NOT_FOUND)
        for corpus in Corpus.objects.filter(elements__in=elements).distinct():
            if Right.Read not in corpus.get_acl_rights(request.user):
                return Response(status=status.HTTP_403_FORBIDDEN)

        # Use sets to ensure IDs are unique, but convert back to a list as session data must be JSON-serializable
        request.session['elements_selected'] = list(
            set(request.session['elements_selected'])
            | set(request.data['ids'])
        )
        request.session.modified = True

        return Response(
            status=status.HTTP_201_CREATED,
            data=ElementSlimSerializer(elements, context={'request': request}, many=True).data,
        )

    def delete(self, request, *args, **kwargs):
        if 'id' not in request.data:
            request.session['elements_selected'] = []
            return Response(status=status.HTTP_204_NO_CONTENT)

        if request.data['id'] not in request.session.get('elements_selected'):
            return Response(status=status.HTTP_400_BAD_REQUEST)

        request.session.get('elements_selected').remove(request.data['id'])
        request.session.modified = True

        return Response(status=status.HTTP_204_NO_CONTENT)


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        return Corpus.objects \
                     .readable(self.request.user, add_stats=True) \
                     .prefetch_related('types')


class CorpusRetrieve(RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        rights = obj.get_acl_rights(request.user)
        if request.method == 'DELETE' and Right.Admin not in rights:
            self.permission_denied(request, message='You do not have admin acces to this corpus.')
        elif request.method != 'GET' and Right.Write not in rights:
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        corpus_delete(instance)


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    List all transcriptions for an element, optionally filtered by type
    """
    serializer_class = TranscriptionSerializer
    pagination_class = TranscriptionsPagination
    openapi_overrides = {
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'required': False,
                'description': 'Transcription type filter',
                'schema': {
                    'type': 'string',
                    'enum': [ts_type.value for ts_type in TranscriptionType],
                }
            },
        ]
    }

    def get_queryset(self):
        queryset = Transcription.objects.filter(
            element_id=self.kwargs['pk'],
            element__corpus__in=Corpus.objects.readable(self.request.user),
        ).prefetch_related('zone__image__server', 'source').order_by('id')
        req_type = self.request.query_params.get('type')
        if req_type:
            try:
                req_type = TranscriptionType(req_type)
                queryset = queryset.filter(type=req_type)
            except ValueError:
                raise ValidationError({'type': 'Not a valid transcription type'})
        return queryset


class ElementRegions(DeprecatedMixin, ListAPIView):
    """
    List all regions on an element.
    """
    openapi_overrides = {
        'operationId': 'ListElementRegions',
        'security': [],
        'tags': ['elements'],
    }
    deprecation_message = \
        'This endpoint has been deprecated. ' \
        'Regions have been migrated to children elements with classifications. Use ListElementChildren to access them.'


class ElementsCreate(CreateAPIView):
    """
    Create a new element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateElement',
        'tags': ['elements'],
    }


class RegionDetails(DeprecatedMixin, RetrieveAPIView):
    """
    Retrieve detailed information about a region.
    """
    openapi_overrides = {
        'operationId': 'RetrieveRegion',
        'security': [],
        'tags': ['ml'],
    }
    deprecation_message = \
        'This endpoint has been deprecated. ' \
        'Regions have been migrated to children elements with classifications. Use ListElementChildren to access them.'


class RegionCreate(DeprecatedMixin, CreateAPIView):
    """
    Create a region on an element.
    """
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'CreateRegion',
        'tags': ['ml'],
    }
    deprecation_message = \
        'This endpoint has been deprecated. ' \
        'Regions have been migrated to children elements with classifications. Use ListElementChildren to access them.'


class RegionBulkCreate(DeprecatedMixin, CreateAPIView, UpdateAPIView):
    """
    Add new regions from a machine learning tool.
    """
    permission_classes = (IsVerified, )
    # Force DRF to ignore PATCH
    http_method_names = ['post', 'put', 'head', 'options', 'trace']
    # Cannot override operationId or description here because openapi_overrides does not handle multiple HTTP methods
    openapi_overrides = {
        'tags': ['ml'],
    }
    deprecation_message = \
        'This endpoint has been deprecated. ' \
        'Regions have been migrated to children elements with classifications. Use ListElementChildren to access them.'


class ElementMetadata(CreateAPIView):
    """
    Create a metadata on an existing element.

    The `date` type allows dates to be parsed and indexed for search APIs. The supported date formats are as follows:

    * `YYYY`: `1995` (recommended)
    * `YYYY-MM`: `1995-02` (recommended)
    * `YYYY-MM-DD`: `1995-02-16` (recommended)
    * `/YYYY(.*)month/`: `1995, february`, with months in French or English (case-insensitive)
    * `YYYY-YYYY`: `1995-1996` (yields a date interval)

    The `reference` type is also indexed in search APIs and allows searching by
    known identifiers or references, or any arbitrary string.
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'operationId': 'CreateMetaData',
        'tags': ['elements']
    }

    def get_queryset(self):
        return Element.objects.filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context


class MetadataEdit(RetrieveUpdateDestroyAPIView):
    """
    Edit an existing element medadata
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'tags': ['elements']
    }

    def get_queryset(self):
        # Filter readable metadata in order to check object permissions
        return MetaData.objects.filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if Right.Write not in obj.element.corpus.get_acl_rights(request.user):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        serializer = self.get_serializer(instance)
        serializer.delete(instance)


class AllowedMetaDataPagination(PageNumberPagination):
    page_size = 200


class CorpusAllowedMetaData(CorpusACLMixin, ListAPIView):
    """
    List allowed metadata (type, name) couples in a corpus
    """
    serializer_class = CorpusAllowedMetaDataSerializer
    pagination_class = AllowedMetaDataPagination
    openapi_overrides = {
        'operationId': 'ListCorpusAllowedMetaDatas',
        'description': 'List (type, name) couples allowed for corpus metadata',
        'tags': ['corpora']
    }

    def get_queryset(self):
        if not self.request:
            return AllowedMetaData.objects.none()
        return AllowedMetaData.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))
